open InterpreterM;;

(** print help in chunks of 25 lines *)
let msg =  
	write_line "LIPS is an interpreter for the lambda-calculus, implemented by Michael Brunner" >>
	write_line "as part of his bachelor project 'lambda-interpreter'." >>
	write_line "LIPS was developed in OCaml using monads to achieve purely functional code." >>
	write_line "" >>
	write_line "It calculates the normal form for every given lambda-term. For example:" >>
  write_line "" >>
	write_line "  > (\\x y.x) a b" >>
	write_line "  NF: a" >>
	write_line "  > (\\f.(\\x.f (x x)) (\\x.f (x x))) (\\y z.z)" >>
	write_line "  NF: \\z.z" >>
	write_line "" >>
	write_line "The syntax of lambda-terms is extended to CoreML-expressions. The full" >>
	write_line "grammar in BNF-notation is:" >>
  write_line "" >>
	write_line "  <term> ::= (<term>) | <var> | <abs> | <app> | <let> | <if> | <con>" >>
	write_line "  <var>  ::= <chr> | <var><chr> | <var><num> | <var>'" >>
	write_line "  <chr>  ::= a | b | ... | z | A | B | ... | Z | _" >>
	write_line "  <num>  ::= 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9" >>
	write_line "  <abs>  ::= \\<vars>.<term>" >>
	write_line "  <vars> ::= <var> | <var> <vars>" >>
	write_line "  <app>  ::= <term> <term>" >>
	write_line "  <let>  ::= let <var> = <term> in <term>" >>
	write_line "  <if>   ::= if <term> then <term> else <term>" >>
	write_line "  <con>  ::= true | false" >>
	write_line "" >>
	write_line "<<PRESS [RETURN] TO CONTINUE>>" >>
	read_line >>
	write_line "For convenience you may also define and use named lambda-terms. To define" >>
  write_line "a named lambda-term type \"!def <var> = <term>\". The given name may from" >>
	write_line "now on be used within lambda-terms and will be replaced by the corresponding" >> 
	write_line "term before evaluating normal forms." >>
  write_line "" >>
  write_line "  > !def K = \\x y.x" >>
  write_line "  variable K added to environment" >>
  write_line "  > K a b" >>
  write_line "  NF: a" >>
  write_line "" >>
	write_line "The command \"!defs\" prints a list of all named lambda-terms." >>
	write_line "" >>
	write_line "To toggle between trace and normal evaluation use the command \"!trace\"." >>
	write_line "In trace-mode all evaluation-steps are displayed and the reduced redexes" >>
	write_line "are highlighted (this may not work in all terminals)." >>
  write_line "" >>
	write_line "  > (\\x y.x) a b" >>
	write_line "  Trace: \027[04m(\\x y.x) a\027[00m b" >>
	write_line "  Trace: \027[04m(\\y.a) b\027[00m" >>
	write_line "  Trace: a" >>
  write_line "  NF: a" >>
  write_line "" >>
	write_line "By default trace-mode is deactivated." >>
	write_line "" >>
	write_line "" >>
	write_line "<<PRESS [RETURN] TO CONTINUE>>" >>
	read_line >>
	write_line "LIPS also supports different evaluation strategies. Currently you" >>
  write_line "can choose amongst by-value, by-name, parallel innermost and parallel" >>
	write_line "outermost reduction. The commands to switch to these strategies are" >> 
	write_line "\"!by_name\", \"!by_value\", \"!parallel_innermost\", and \"!parallel_outermost\"." >>
  write_line "" >>
  write_line "  > (\\x y.x) z ((\\x y.y) z ((\\x.x) z))" >>
  write_line "  Trace: \027[04m(\\x y.x) z\027[00m ((\\x y.y) z ((\\x.x) z))" >>
  write_line "  Trace: \027[04m(\\y.z) ((\\x y.y) z ((\\x.x) z))\027[00m" >>
	write_line "  Trace: z" >>
	write_line "  NF: z" >>
	write_line "  > !parallel_outermost" >>
	write_line "  Evaluation-strategy changed to parallel outermost" >>
	write_line "  > (\\x y.x) z ((\\x y.y) z ((\\x.x) z))" >>
	write_line "  Trace: \027[04m(\\x y.x) z\027[00m (\027[04m(\\x y.y) z\027[00m (\027[04m(\\x.x) z\027[00m))" >>
	write_line "  Trace: \027[04m(\\y.z) ((\\y.y) z)\027[00m" >>
	write_line "  Trace: z" >>
	write_line "  NF: z" >>
	write_line "  >" >>
  write_line "" >>
	write_line "By default LIPS uses by-name (i.e., leftmost outermost) reduction." >>
	write_line "" >>
	write_line "" >>
	write_line "" >>
	write_line "" >>
	write_line "" >>
	write_line "<<PRESS [RETURN] TO CONTINUE>>" >>
	read_line >>
	write_line "LIPS also provides a type-system, which can be activated by entering" >>
	write_line "the command \"!typing\". When activated, only well typed lambda-terms" >>
	write_line "are accepted and upon normal-form-evaluation the inferred type of the" >>
	write_line "given expression is shown. " >>
  write_line "" >>
	write_line "  > \\x y.x" >>
	write_line "  NF: \\x y.x : 'a -> 'b -> 'a" >>
	write_line "  > \\x.if x then false else true" >>
	write_line "  NF: \\x.if x then false else true : bool -> bool" >>
  write_line "" >>
	write_line "Type-checking is also supported. Input \"<term> : <type>\" to perform a" >>
	write_line "type-check. In this case the given term is not evaluated." >>
	write_line "" >>
	write_line "It is possible to define (or redefine) the type of any named lambda-term" >>
	write_line "by \"!type <var> : <type>\". It is also possible to denote the type of a " >>
	write_line "named lambda-term directly with \"!def <var> : <type> = <term>\"." >>
	write_line "The command \"!types\" shows all type-definitions." >>
	write_line "" >>
	write_line "The grammar for types in BNF notation is" >>
  write_line "" >>
	write_line "  <type> ::= <tvar> | <tcon> | <tcon>(<tpar>) | <type> -> <type>" >>
	write_line "  <tvar> ::= '<chr>" >>
	write_line "  <tcon> ::= <chr> | <tcon><chr>" >>
	write_line "  <tpar> ::= <type> | <type>,<tpar>" >>
	write_line "" >>
	write_line "<<PRESS [RETURN] TO CONTINUE>>" >>
	read_line >>
	write_line "The commands \"!run <file>\" and load \"!load <file>\" execute the given" >>
	write_line "file silently (just show number of lines processed) or verbosely." >>
	write_line "" >>
	write_line "Further, LIPS provides basic session-management allowing to export " >>
	write_line "arbitrary parts of an interpreter session to an external file. A basic " >>
	write_line "template system provides means necessary to customize the format of " >>
	write_line "the exported session. The command \"!template <filename>\" loads a template" >>
	write_line "from the given file. By default a template is used, which exports the session" >>
	write_line "in ASCII-format similar to the appearance in LIPS." >>
	write_line "To start session-recording the command \"!record\" has to be entered. From " >>
	write_line "this moment on, all normal form evaluations are cached. To export the session" >>
	write_line "to a file, the command \"!flush <filename>\" has to be used. This exports the " >>
	write_line "collected session-data to the given file, using the current template." >>
	write_line "" >>
	write_line "To exit LIPS type the command \"!quit\"."
;;


