(* Copyright 2008, 2009 Michael Brunner *)

(* This file is part of LIPS.

  LIPS is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as 
	published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

  LIPS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
	License along with LIPS.  If not, see 
	<http://www.gnu.org/licenses/>. *)

open Monad;;


module Make (M: MONAD_MINIMAL) = struct 
	type 'a t = 'a M.t
	let (>>=) = M.bind
	let (>>) m n = m >>= fun _ -> n 
	let return = M.unit
end;;


module ExceptionT (M: MONAD) = struct 
	type 'a exc = Exception of string | Result of 'a;;
	module ExceptionM = struct
	  	type 'a t = ('a exc) M.t;;
			let unit x = M.return (Result x);;
			let bind m f = M.(>>=) m (function
				| Result r -> f r 
				| Exception e -> M.return (Exception e));;
	end
  include Make (ExceptionM);;
  let raise e = M.return (Exception e);;
  let trywith m h = M.(>>=) m (function Exception e -> h e | Result r -> return r);;
	let run m = match M.run m with Exception (e) -> failwith e | Result (r) -> r;;	
  let lift m = M.(>>=) m (fun x -> return x);; 
end;;


module StateT (S: TYPEWRAPPER) (M: MONAD) = struct 
	type state = S.t;;
	module StateM = struct
	  	type 'a t = state -> ('a * state) M.t;;
			let unit x = fun s -> M.return (x, s);;
			let bind m f = fun s -> M.(>>=) (m s) (fun (x, s) -> f x s);;
	end
	include Make (StateM);;
  let read_state = fun s -> M.return (s, s);;
  let write_state x = fun s -> M.return ((), x);;
  let run m = fst (M.run (m S.init));;
  let run_with m s = fst (M.run (m s));;
  let lift m = fun s -> M.(>>=) m (fun x -> M.return (x, s));;
end;;


