(* Copyright 2008, 2009 Michael Brunner *)

(* This file is part of LIPS.

  LIPS is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as 
	published by the Free Software Foundation, either version 3 of 
	the License, or (at your option) any later version.

  LIPS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
	License along with LIPS.  If not, see 
	<http://www.gnu.org/licenses/>. *)

type entry = Nfe of (CoreML.t * (CoreML.t * CoreML.pos list) list * CoreML.t) 
        | TNfe of (CoreML.t * (CoreML.t * CoreML.pos list) list * CoreML.t * Type.t);;

type t = entry list;;

type template = {
	simple_output: bool; (* define to use simple-output for this template  *)
	
  doc_prefix: Strng.t; (* document prefix *)
	doc_suffix: Strng.t; (* document suffix *)
	eval_prefix: Strng.t; (* evaluation prefix *)
	eval_suffix: Strng.t; (* evaluation suffix *)
	
	beta: Strng.t; (* symbol for beta-step *)
	beta_start: Strng.t; (* symbol for first reduction-step, replacement of named lambda terms *)
	beta_nf: Strng.t; (* symbol for last beta-step *)
	mark_prefix: Strng.t; (* prefix for marked coreml-expressions, redex-highlighting *)
	mark_suffix: Strng.t; (* suffix for marked coreml-expressions, redex-highlighting *)
	
	lpar: Strng.t; (* symbol for left parenthesis *)
	rpar: Strng.t; (* symbol for right parenthesis *)
	var_prefix: Strng.t; (* prefix for coreml-variables *)
	vari_prefix: Strng.t; (* prefix for variable-index *)
	vari_suffix: Strng.t; (* suffix for variable-index *)
	var_suffix: Strng.t; (* suffiy for coreml-variables *)
	abs_prefix: Strng.t; (* abstraction-prefix, lambda *)
	abs_sep: Strng.t; (* abstraction-variable-seperator, nested abstraction *)
	abs_dot: Strng.t; (* abstraction-seperator, dot *)
	abs_suffix: Strng.t; (* abstraction suffix *)
	app_prefix: Strng.t; (* application-prefix *)
	app_sep: Strng.t; (* application-seperator, space *)
	app_suffix: Strng.t; (* application-suffix *)
	let_prefix: Strng.t; (* prefix for let-binding, let *)
	let_map: Strng.t; (* map-symbol for let-binding, = *)
	let_in: Strng.t; (* seperator for let-binding, in *)
	let_suffix: Strng.t; (* suffix for let-binding *)
	if_prefix: Strng.t; (* prefix for if-clause, if *)
	if_then: Strng.t; (* seperator for then-branch, then *)
	if_else: Strng.t; (* seperator for else-branch, else *)
	if_suffix: Strng.t;	(* suffix for if-clause *)
	
	type_sep: Strng.t; (* seperator for type-definition, : *)
	type_prefix: Strng.t; (* prefix for type *)
	tvar_prefix: Strng.t; (* prefix for type-variable, 'a *)
	tvar_suffix: Strng.t; (* suffix for type-variable *)
	tarrow: Strng.t; (* symbol for arrow-type, -> *)
	tcons_prefix: Strng.t; (* prefix for type-constructor *)
	tcons_suffix: Strng.t; (* suffix for type-constructor *)
	tlpar: Strng.t; (* left parenthesis for n-ary type construtors *)
	trpar: Strng.t; (* right parenthesis for n-ary type constructors *)
	tcomma: Strng.t; (* comma for n-ary type-constructors *)
	type_suffix: Strng.t; (* suffix for type-definiton *)
};;



(** default -template *)
let default = {
	simple_output = true;
  doc_prefix = [];	doc_suffix = [];	eval_prefix = ['>'];	eval_suffix = ['\n';'\n'];	
	beta = Strng.of_string "\nTrace: ";	beta_start = Strng.of_string "\nTrace: "; beta_nf = Strng.of_string "\nNF: ";
	mark_prefix = Strng.of_string "[";	mark_suffix = Strng.of_string "]";	lpar = ['('];	rpar = [')'];
	var_prefix = [];	vari_prefix = ['_'];	vari_suffix = [];	var_suffix = [];
	abs_prefix = ['\\'];	abs_sep = [' '];	abs_dot = ['.'];	abs_suffix = [];	
	app_prefix = [];	app_sep = [' '];	app_suffix = [];	
	let_prefix = Strng.of_string "let ";	let_map = Strng.of_string " = ";	let_in = Strng.of_string " in ";	let_suffix = [];
	if_prefix = Strng.of_string "if ";	if_then = Strng.of_string " then ";	if_else = Strng.of_string " else ";	if_suffix = [];	
	type_sep = [' ';':';' '];	type_prefix = [];	tvar_prefix = ['\'';'a';'_'];	tvar_suffix = [];
	tarrow = ['-';'>'];	tcons_prefix = [];	tcons_suffix = [];
	tlpar = ['('];	trpar = [')'];	tcomma = [',';' '];	type_suffix = [];
};;

open Parser;;
(** parser for tempate-file *)
let rec tmplp tmpl s = (tempt (tmplp' tmpl >>= fun tmpl' -> tmplp tmpl') <|> return tmpl) s 
and tmplp' tmpl s = (
	(tempt (string "simple_output:") >> ((string "true" >> newline >> return {tmpl with simple_output = true}) <|> (string "false" >> newline >> return {tmpl with simple_output = false}))
	<|> tempt (string "doc_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with doc_prefix = cs})
	<|> tempt (string "doc_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with doc_suffix = cs})
	<|> tempt (string "eval_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with eval_prefix = cs})
	<|> tempt (string "eval_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with eval_suffix = cs})
	<|> tempt (string "beta:" >> umask >>= fun cs -> newline >> return {tmpl with beta = cs})
	<|> tempt (string "beta_start:" >> umask >>= fun cs -> newline >> return {tmpl with beta_start = cs}) 
	<|> tempt (string "beta_nf:" >> umask >>= fun cs -> newline >> return {tmpl with beta_nf = cs}) 
	<|> tempt (string "mark_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with mark_prefix = cs}) 
	<|> tempt (string "mark_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with mark_suffix = cs}) 
	<|> tempt (string "lpar:" >> umask >>= fun cs -> newline >> return {tmpl with lpar = cs}) 
	<|> tempt (string "rpar:" >> umask >>= fun cs -> newline >> return {tmpl with rpar = cs}) 
	<|> tempt (string "var_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with var_prefix = cs}) 
	<|> tempt (string "vari_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with vari_prefix = cs}) 
	<|> tempt (string "vari_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with vari_suffix = cs}) 
	<|> tempt (string "var_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with var_suffix = cs}) 
	<|> tempt (string "abs_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with abs_prefix = cs}) 
	<|> tempt (string "abs_sep:" >> umask >>= fun cs -> newline >> return {tmpl with abs_sep = cs}) 
	<|> tempt (string "abs_dot:" >> umask >>= fun cs -> newline >> return {tmpl with abs_dot = cs}) 
	<|> tempt (string "abs_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with abs_suffix = cs}) 
	<|> tempt (string "app_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with app_prefix = cs}) 
	<|> tempt (string "app_sep:" >> umask >>= fun cs -> newline >> return {tmpl with app_sep = cs}) 
	<|> tempt (string "app_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with app_suffix = cs}) 
	<|> tempt (string "let_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with let_prefix = cs}) 
	<|> tempt (string "let_in:" >> umask >>= fun cs -> newline >> return {tmpl with let_in = cs}) 
	<|> tempt (string "let_map:" >> umask >>= fun cs -> newline >> return {tmpl with let_map = cs}) 
	<|> tempt (string "let_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with let_suffix = cs}) 
	<|> tempt (string "if_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with if_prefix = cs}) 
	<|> tempt (string "if_then:" >> umask >>= fun cs -> newline >> return {tmpl with if_then = cs}) 
	<|> tempt (string "if_else:" >> umask >>= fun cs -> newline >> return {tmpl with if_else = cs}) 
	<|> tempt (string "if_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with if_suffix = cs})
	<|> tempt (string "type_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with type_prefix = cs})
	<|> tempt (string "type_sep:" >> umask >>= fun cs -> newline >> return {tmpl with type_sep = cs})
	<|> tempt (string "type_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with type_suffix = cs})
	<|> tempt (string "tvar_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with tvar_prefix = cs})
	<|> tempt (string "tvar_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with tvar_suffix = cs})
	<|> tempt (string "tarrow:" >> umask >>= fun cs -> newline >> return {tmpl with tarrow = cs})
	<|> tempt (string "tcons_prefix:" >> umask >>= fun cs -> newline >> return {tmpl with tcons_prefix = cs})
	<|> tempt (string "tcons_suffix:" >> umask >>= fun cs -> newline >> return {tmpl with tcons_suffix = cs})
	<|> tempt (string "tlpar:" >> umask >>= fun cs -> newline >> return {tmpl with tlpar = cs})
	<|> tempt (string "trpar:" >> umask >>= fun cs -> newline >> return {tmpl with trpar = cs})
	<|> tempt (string "tcomma:" >> umask >>= fun cs -> newline >> return {tmpl with tcomma = cs})
	)) s
and newline = many (one_of "\n\r\t ")
and umask s = (char '"' >> many_until ((tempt (string "\\\"" >> return '"')) <|> any_char) (char '"') >>= fun cs -> return cs) s;;







(** string converter for coreml-expressions, not using conventions *)
let coreml_to_strng' tmpl ps t = 
  let rec get_nested_abs_vars = function 
    | CoreML.Abs (v, l) -> v :: get_nested_abs_vars l
    | _ -> [] in
  let rec get_nested_abs_term = function 
    | CoreML.Abs (v, l) -> get_nested_abs_term l
    | l -> l in
  let var2strng (id, i) =
    tmpl.var_prefix @ (if i <= 0 then id else id @ tmpl.vari_prefix @ (Strng.of_int i) @ tmpl.vari_suffix) @ tmpl.var_suffix in
  let rec vars2strng = function 
    | [] -> []
    | x::[] -> (var2strng x)
    | x::xs -> (var2strng x) @ tmpl.abs_sep @ (vars2strng xs) in
  let cons2strng = function 
	  | CoreML.True -> (Strng.of_string "true")
	  | CoreML.False -> (Strng.of_string "false") in
	let rec to_strng pos t = if Lst.mem pos ps then tmpl.mark_prefix @ to_strng' pos t @ tmpl.mark_suffix else to_strng' pos t 
	and to_strng' pos = function
		| CoreML.Var v -> var2strng v
		| CoreML.Abs (v, t) -> tmpl.lpar @ tmpl.abs_prefix @ (var2strng v) @ tmpl.abs_dot @ (to_strng (pos@[CoreML.Left]) t) @ tmpl.rpar
		| CoreML.App (t1, t2) -> tmpl.lpar @ (to_strng (pos@[CoreML.Left]) t1) @ tmpl.app_sep @ (to_strng (pos@[CoreML.Right]) t2) @ tmpl.rpar
		| CoreML.Let (v,t, b) -> tmpl.lpar @ tmpl.let_prefix @ (var2strng v) @ tmpl.let_map @ (to_strng (pos@[CoreML.Left]) t) @ tmpl.let_in @ (to_strng (pos@[CoreML.Right]) b) @ tmpl.let_suffix @ tmpl.rpar
		| CoreML.If (c, t1, t2) -> tmpl.lpar @ tmpl.if_prefix @ (to_strng (pos@[CoreML.Left]) c) @ tmpl.if_then @ (to_strng (pos@[CoreML.Middle]) t1) @ tmpl.if_else @ (to_strng (pos@[CoreML.Right]) t2) @ tmpl.if_suffix @ tmpl.rpar
		| CoreML.Cons c -> cons2strng c
	in to_strng [] t;;

(** string converter for coreml-expressions using conventions, simple output*)
let coreml_to_strng tmpl ps t = 
  let rec get_nested_abs_vars = function 
    | CoreML.Abs (v, l) -> v :: get_nested_abs_vars l
    | _ -> [] in
  let rec get_nested_abs_term = function 
    | CoreML.Abs (v, l) -> get_nested_abs_term l
    | l -> l in
  let var2strng (id, i) =
    tmpl.var_prefix @ (if i <= 0 then id else id @ tmpl.vari_prefix @ (Strng.of_int i) @ tmpl.vari_suffix) @ tmpl.var_suffix in
  let rec vars2strng = function 
    | [] -> []
    | x::[] -> (var2strng x)
    | x::xs -> (var2strng x) @ tmpl.abs_sep @ (vars2strng xs) in
  let cons2strng = function 
	  | CoreML.True -> (Strng.of_string "true")
	  | CoreML.False -> (Strng.of_string "false") in
	let rec to_strng pos t = if Lst.mem pos ps then tmpl.mark_prefix @ to_strng' pos t @ tmpl.mark_suffix else to_strng' pos t 
	and to_strng' pos = function
		| CoreML.Var v -> var2strng v
		| CoreML.Abs (v, t) -> let vs = v::get_nested_abs_vars t 
										and ls = get_nested_abs_term t in 
										tmpl.abs_prefix @ (vars2strng vs) @ tmpl.abs_dot @ (to_strng (pos @ (Lst.init CoreML.Left (Lst.length vs))) ls)
(*		| CoreML.App (t1, CoreML.App(t2, t3)) -> (to_strng (pos@[CoreML.Left]) t1) @ tmpl.app_sep @ tmpl.lpar @ (to_strng (pos@[CoreML.Right]) (CoreML.App (t2, t3))) @ tmpl.rpar*)
		| CoreML.App (CoreML.Abs (v1, t1), CoreML.Abs (v2, t2)) -> tmpl.lpar @ (to_strng (pos@[CoreML.Left]) (CoreML.Abs (v1, t1))) @ tmpl.rpar @ tmpl.app_sep @ tmpl.lpar @ (to_strng (pos@[CoreML.Right]) (CoreML.Abs (v2, t2))) @ tmpl.rpar
		| CoreML.App (CoreML.Abs (v1, t1), CoreML.App(t2, t3)) -> tmpl.lpar @ (to_strng (pos@[CoreML.Left]) (CoreML.Abs (v1, t1))) @ tmpl.rpar @ tmpl.app_sep @ tmpl.lpar @ (to_strng (pos@[CoreML.Right]) (CoreML.App (t2, t3))) @ tmpl.rpar
		| CoreML.App (CoreML.Abs (v1, t1), t2) -> tmpl.lpar @ (to_strng (pos@[CoreML.Left]) (CoreML.Abs (v1, t1))) @ tmpl.rpar @ tmpl.app_sep @ (to_strng (pos@[CoreML.Right]) t2)
		| CoreML.App (t1, CoreML.App(t2, t3)) -> (to_strng (pos@[CoreML.Left]) t1) @ tmpl.app_sep @ tmpl.lpar @ (to_strng (pos@[CoreML.Right]) (CoreML.App (t2, t3))) @ tmpl.rpar
		| CoreML.App (t1, CoreML.Abs (v2, t2)) -> (to_strng (pos@[CoreML.Left]) t1) @ tmpl.app_sep @ tmpl.lpar @ (to_strng (pos@[CoreML.Right]) (CoreML.Abs (v2, t2))) @ tmpl.rpar
		| CoreML.App (t1, t2) -> (to_strng (pos@[CoreML.Left]) t1) @ tmpl.app_sep @ (to_strng (pos@[CoreML.Right]) t2)
		| CoreML.Let (v,t, b) -> tmpl.let_prefix @ (var2strng v) @ tmpl.let_map @ (to_strng (pos@[CoreML.Left]) t) @ tmpl.let_in @ (to_strng (pos@[CoreML.Right]) b) @ tmpl.let_suffix
		| CoreML.If (c, t1, t2) -> tmpl.if_prefix @ (to_strng (pos@[CoreML.Left]) c) @ tmpl.if_then @ (to_strng (pos@[CoreML.Middle]) t1) @ tmpl.if_else @ (to_strng (pos@[CoreML.Right]) t2) @ tmpl.if_suffix
		| CoreML.Cons c -> cons2strng c
	in to_strng [] t;;

(** string converter for types *)
let type_to_strng tmpl typ = 
	let rec type_to_strng = function 
		| Type.Var i -> tmpl.tvar_prefix @ Strng.of_string (string_of_int i) @ tmpl.tvar_suffix
		| Type.Arrow (Type.Arrow (t1, t1'), t2) -> tmpl.tlpar @ (type_to_strng t1) @ tmpl.tarrow @ (type_to_strng t1') @ tmpl.trpar @ tmpl.tarrow @ (type_to_strng t2)
		| Type.Arrow (t1, t2) -> (type_to_strng t1) @ tmpl.tarrow @ (type_to_strng t2)
		| Type.TCons (s, []) -> tmpl.tcons_prefix @ s @ tmpl.tcons_suffix
		| Type.TCons (s, ts) -> tmpl.tcons_prefix @ s @ tmpl.tcons_suffix @ tmpl.tlpar @ (Lst.fold1 (fun x y -> x @ tmpl.tcomma @ y) (Lst.map type_to_strng ts)) @ tmpl.trpar 
	in type_to_strng typ;;

(** string converter for session *)
let to_strng tmpl session = 
	let strng_converter = if tmpl.simple_output then coreml_to_strng else coreml_to_strng' in
	let to_strng_entry = function 
		| Nfe (i, t, r) -> tmpl.eval_prefix @ (strng_converter tmpl [] i) @ (if t = [] then [] else tmpl.beta_start @ Lst.fold1 (fun x xs -> x @ tmpl.beta @ xs) (Lst.map (fun (l, ps) -> strng_converter tmpl ps l) t)) @ tmpl.beta_nf @ (strng_converter tmpl [] r) @ tmpl.eval_suffix
		| TNfe (i, t, r, t') -> tmpl.eval_prefix @ (strng_converter tmpl [] i) @ (if t = [] then [] else tmpl.beta_start @ Lst.fold1 (fun x xs -> x @ tmpl.beta @ xs) (Lst.map (fun (l, ps) -> strng_converter tmpl ps l) t)) @ tmpl.beta_nf @ (strng_converter tmpl [] r) @ tmpl.type_sep @ tmpl.type_prefix @ (type_to_strng tmpl t') @ tmpl.type_suffix @ tmpl.eval_suffix in
	let rec to_strng = function 
		| [] -> []
		| (x::xs) -> (to_strng_entry x) @ (to_strng xs) in
	tmpl.doc_prefix @ (to_strng session) @ tmpl.doc_suffix;;
